#ifndef BASE_ATOMIC_H
#define BASE_ATOMIC_H

#include <atomic>
#include <functional>
#include <type_traits>

namespace base {
   template<typename T>
   inline typename std::enable_if<sizeof(T) == sizeof(std::atomic<T>) && alignof(T) == alignof(std::atomic<T>), std::atomic<T>&>::type as_atomic(T& v)
   {
      return reinterpret_cast<std::atomic<T>&>(v);
   }


   template<typename T, typename PRED>
   inline T set_min(std::atomic<T>& destino, const T& valor, const PRED& pred)
   {
      T actual = destino.load( );

      while (pred(valor, actual)) {
         if (destino.compare_exchange_weak(actual, valor)) {
            return valor;
         }
      }

      return actual;
   }

   template<typename T>
   inline T set_min(std::atomic<T>& destino, const T& valor)
   {
      return set_min(destino, valor, std::less<T>( ));
   }


   template<typename T, typename PRED>
   inline T set_max(std::atomic<T>& destino, const T& valor, const PRED& pred)
   {
      T actual = destino.load( );

      while (pred(actual, valor)) {
         if (destino.compare_exchange_weak(actual, valor)) {
            return valor;
         }
      }

      return actual;
   }

   template<typename T>
   inline T set_max(std::atomic<T>& destino, const T& valor)
   {
      return set_max(destino, valor, std::less<T>( ));
   }
}

#endif
