#ifndef BASE_BLOCKING_QUEUE_H
#define BASE_BLOCKING_QUEUE_H

#include <condition_variable>
#include <mutex>
#include <utility>

namespace base {
   template<typename T, bool (T::*try_pop)(typename T::value_type&) = T::try_pop>
   struct blocking_queue_adaptor {
      blocking_queue_adaptor(T& c)
      : cola_(c)
      {
      }

      void push(typename T::value_type&& v)
      {
         cola_.push(std::move(v));
         cv_.notify_one( );
      }

      void push(const typename T::value_type& v)
      {
         return push(typename T::value_type(v));
      }

      template<typename... PARAMS>
      void emplace(PARAMS&&... p)
      {
         return push(typename T::value_type(std::forward<PARAMS>(p)...));
      }

      void wait_pop(typename T::value_type& v)
      {
         std::unique_lock<std::mutex> centinela_(lock_);

         while (!(cola_.*try_pop)(v)) {
            cv_.wait(centinela_);
         }
      }

      T& get_queue( )
      {
         return cola_;
      }

      const T& get_queue( ) const
      {
         return cola_;
      }

   private:
      T& cola_;

      std::mutex lock_;
      std::condition_variable cv_;
   };
}

#endif
