#ifndef BASE_CHAR_H
#define BASE_CHAR_H

#include <cctype>
#include <cstddef>
#include <cstdlib>
#include <cstring>
#include <cwchar>
#include <cwctype>
#include <utility>

namespace base {
   inline std::size_t strlen(const char* s)
   {
      return std::strlen(s);
   }

   inline std::size_t strlen(const wchar_t* s)
   {
      return std::wcslen(s);
   }

   inline bool isalpha(char c)
   {
      return std::isalpha(static_cast<unsigned char>(c));
   }

   inline bool isalpha(wchar_t c)
   {
      return std::iswalpha(c);
   }

   inline bool isalnum(char c)
   {
      return std::isalnum(static_cast<unsigned char>(c));
   }

   inline bool isalnum(wchar_t c)
   {
      return std::iswalnum(c);
   }

   inline bool isdigit(char c)
   {
      return std::isdigit(static_cast<unsigned char>(c));
   }

   inline bool isdigit(wchar_t c)
   {
      return std::iswdigit(c);
   }

   inline bool isspace(char c)
   {
      return std::isspace(static_cast<unsigned char>(c));
   }

   inline bool isspace(wchar_t c)
   {
      return std::iswspace(c);
   }

   inline bool ispunct(char c)
   {
      return std::ispunct(static_cast<unsigned char>(c));
   }

   inline bool ispunct(wchar_t c)
   {
      return std::iswpunct(c);
   }

   inline char tolower(char c)
   {
      return std::tolower(static_cast<unsigned char>(c));
   }

   inline wchar_t tolower(wchar_t c)
   {
      return std::towlower(c);
   }

   inline char toupper(char c)
   {
      return std::toupper(static_cast<unsigned char>(c));
   }

   inline wchar_t toupper(wchar_t c)
   {
      return std::towupper(c);
   }


   inline int chrtod(char c, unsigned base)
   {
      char temp[] = { c, '\0' };
      char* fin;

      auto res = std::strtol(temp, &fin, base);
      return (fin > temp ? res : -1);
   }

   template<typename T, typename FI>
   inline std::pair<T, FI> strtod(FI i, FI f, unsigned base)
   {
      T res = 0;

      while (i < f) {
         auto d = chrtod(*i, base);

         if (d < 0) {
            break;
         }

         res *= base;
         res += d;

         ++i;
      }

      return { res, i };
   }
}

#endif
