#ifndef BASE_FILE_H
#define BASE_FILE_H

#include "intrinsic.h"
#include <cstddef>
#include <cstdio>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>

namespace base {
   class file {
   public:
      file( )
      : f_(nullptr)
      {
      }
      
      file(const char* c, const char* m)
      : f_(std::fopen(c, m))
      {
      }
      
   #ifdef _WIN32
      file(const wchar_t* c, const wchar_t* m)
      : f_(::_wfopen(c, m))
      {
      }
   #endif
      
      file(const file&) = delete;
      
      file(file&& f)
      : f_(f.f_)
      {
         f.f_ = nullptr;
      }
      
      ~file( )
      {
         if (f_ != nullptr) {
            std::fclose(f_);
         }
      }
      
      void operator=(const file&) = delete;
      
      void operator=(file&& f)
      {
         move(*this, f);
      }
      
      bool is_open( ) const
      {
         return f_;
      }
      
      bool close( )
      {
         return std::fclose(f_);
      }
            
      bool flush( )
      {
         return std::fflush(f_) == 0;
      }
            
      FILE* get( )
      {
         return f_;
      }
      
   private:
      FILE* f_;
   };
   
   #ifdef _WIN32
      using file_attributes = struct ::_stat;
   
      inline bool stat(const char* ruta, file_attributes* buffer)
      {
         return (::_stat(ruta, buffer) == 0);
      }
      
      inline bool stat(const wchar_t* ruta, file_attributes* buffer)
      {
         return (::_wstat(ruta, buffer) == 0);
      }
   #else
		using file_attributes = struct ::stat;
      
      inline bool stat(const char* ruta, file_attributes* buffer)
      {
         return (::stat(ruta, buffer) == 0);
      }
   #endif
   
   std::ptrdiff_t file_size(FILE* f)
   {
      static_assert(sizeof(std::ptrdiff_t) >= sizeof(::off_t), "base::file_size");
      
      struct ::stat buffer;
      return (::fstat(fileno(f), &buffer) == 0 ? buffer.st_size : -1);
   }
}

#endif
