#ifndef BASE_UTILITY_H
#define BASE_UTILITY_H

#include <utility>

namespace base {
   template<typename F>
   class exit_guard {
   public:
      exit_guard(F&& v)
      : func_(std::move(v))
      {
      }

      exit_guard(const F& v)
      : func_(v)
      {
      }

      ~exit_guard( )
      {
         func_( );
      }

   private:
      F func_;
   };

   template<typename F>
   inline exit_guard<F> make_exit_guard(F&& v)
   {
      return std::move(v);
   }

   template<typename F>
   inline exit_guard<F> make_exit_guard(const F& v)
   {
      return v;
   }
}

#endif
