//#define TBB_USE_DEBUG 1
//#define TBB_USE_THREADING_TOOLS 1

#include <tbb/task_scheduler_init.h>
#include <bitset>
#include <fstream>
#include <iostream>
#include <map>
#include <set>
#include <sstream>

#include "grevlex.h"
#include "monomial.h"
#include "polynomial.h"
#include "polynomial_io.h"
#include "sparse_polynomial.h"
#include "base/functional.h"
#include <tbb/concurrent_vector.h>
#include <tbb/parallel_for.h>
#include <tbb/parallel_for_each.h>
#include <tbb/parallel_sort.h>
#include <tbb/partitioner.h>
#include <tbb/task.h>
#include <algorithm>
#include <array>
#include <atomic>
#include <chrono>
#include <cmath>
#include <cstddef>
#include <cstdint>
#include <iostream>
#include <shared_mutex>
#include <utility>
#include <tuple>

template<std::uint32_t N, typename II, typename P>
inline void lee_polinomio(P& p, II ai, II af, const grevlex_cache<N>& conv)
{
   while (ai != af) {
      p.insert_erase(conv.index_of(*ai++));
   }
}

template<std::uint32_t N, typename OI, typename P>
inline OI escribe_polinomio(const P& p, OI oi, const grevlex_cache<N>& conv)
{
   p.visit([&](auto i) {
      *oi++ = conv.monomial_of(i);
   });

   return oi;
}

template<typename F>
inline void visita_disperso(const sparse_polynomial& origen, F&& f)
{
   origen.leaf_visit(f);
}

template<typename F, std::uint32_t N>
inline void visita_producto_disperso(const sparse_polynomial& origen, const monomial<N>& factor, F&& f, const grevlex_cache<N>& conv)
{
   return (factor.none( ) ? visita_disperso(origen, std::forward<F>(f)) : origen.visit([&](auto indice) {
      f(conv.index_of(factor * conv.monomial_of(indice)));
   }));
}

template<std::uint32_t N>
inline monomial<N> genera_factor(std::uint32_t i)
{
   monomial<N> res;
   res.insert(i);

   return res;
}

template<std::uint32_t N>
class piramide_polinomial {
   struct registro {
      registro( )
      : reductor(-1)
      {
      }

      std::uint32_t clave : 31;
      std::uint32_t denso : 1;
      std::uint32_t reductor;
      sparse_polynomial polinomio;
   };

public:
   struct reducible {
      reducible(std::uint32_t n)
      : polinomio(n), indice_lider(-1)
      {
      }

      polynomial polinomio;
      std::uint32_t indice_lider;
   };

   struct reductor {
      reductor(std::uint32_t n)
      : polinomio(n), lista(base::dynamic_create_array<std::uint32_t>(std::allocator<std::uint32_t>( ), list_capacity(n))), fin_lista(nullptr), representante(nullptr)
      {
      }

      reductor(reductor&& r)
      : polinomio(std::move(r.polinomio)), lista(std::move(r.lista)), fin_lista(r.fin_lista.load(std::memory_order_relaxed)), representante(r.representante.load(std::memory_order_relaxed))
      {
      }

      static std::uint32_t list_capacity(std::uint32_t n)
      {
         return base::ceil_division(n, CHAR_BIT * (sizeof(std::uint32_t) + sizeof(vector_type)));
      }

      polynomial polinomio;
      std::unique_ptr<std::uint32_t[]> lista;
      std::atomic<std::uint32_t*> fin_lista;
      std::atomic<const polynomial*> representante;
   };

   piramide_polinomial( )
   : clave_(0), grado_(-1)
   {
   }

   std::uint32_t degree( ) const
   {
      return grado_;
   }

   std::uint32_t size( ) const
   {
      return datos_.size( );
   }

   std::pair<std::uint32_t, bool> get_status(std::uint32_t i) const
   {
      return { datos_[i].clave, datos_[i].reductor == i };
   }

   const sparse_polynomial& operator[](std::uint32_t i) const
   {
      return datos_[i].polinomio;
   }

   void increase_degree(const grevlex_cache<N>& conv)
   {
      datos_.resize(conv.population_up_to(++grado_));
   }

   void preprocess(tbb::concurrent_vector<std::pair<std::uint32_t, std::uint32_t>>& multiplos_pendientes, const grevlex_cache<N>& conv)
   {
      tbb::parallel_sort(multiplos_pendientes.begin( ), multiplos_pendientes.end( ));
      multiplos_pendientes.resize(std::unique(multiplos_pendientes.begin( ), multiplos_pendientes.end( )) - multiplos_pendientes.begin( ));
      multiplos_pendientes.resize(std::remove_if(multiplos_pendientes.begin( ), multiplos_pendientes.end( ), [](const auto& producto) {
         return producto.second == 0;
      }) - multiplos_pendientes.begin( ));
std::cerr << multiplos_pendientes.size( ) << " multiplos pendientes sin duplicados\n";
      tbb::parallel_for(std::uint32_t(0), std::uint32_t(datos_.size( )), [&, this](auto indice) {
         auto& datos = datos_[indice];
         datos.reductor = (datos.polinomio.empty( ) ? std::uint32_t(-1) : indice);
         datos.denso &= (datos.reductor != std::uint32_t(-1));
      });

      tbb::concurrent_vector<std::pair<std::uint32_t, std::uint32_t>> multiplos_finales;
      std::for_each(multiplos_pendientes.begin( ), multiplos_pendientes.end( ), [&, this](const auto& producto) {
         auto& datos_multiplo = datos_[conv.index_of(least_common_multiple(conv.monomial_of(producto.first), conv.monomial_of(producto.second)))];
         if (datos_multiplo.reductor == std::uint32_t(-1)) {
            datos_multiplo.reductor = producto.first;
         }
         else {
            multiplos_finales.push_back(producto);
         }
      });
std::cerr << multiplos_finales.size( ) << " multiplos pendientes finales\n";
      multiplos_pendientes = std::move(multiplos_finales);
std::cerr << "a preparar los reductores de las filas faltantes\n";
      tbb::parallel_for(std::uint32_t(0), std::uint32_t(datos_.size( )), [&, this](auto indice) {
         auto& datos = datos_[indice];
         if (datos.reductor == std::uint32_t(-1)) {
            datos.reductor = this->busca_reductor_(indice, conv);
         }
      });
   }

   template<typename RI1, typename RI2>
   void import(RI1 pi, RI1 pf, RI2 ri, RI2 rf, tbb::concurrent_vector<std::uint32_t>& indices, const grevlex_cache<N>& conv)
   {
      auto reducibles = base::make_range(pi, pf);
      auto reductores = base::make_range(ri, rf);
      auto fin_actual = std::uint32_t(datos_.size( ));

      tbb::parallel_for_each(reductores, [&, this](auto& reductor) {
         reductor.polinomio.clear(datos_.size( ));
         reductor.representante.store(nullptr, std::memory_order_relaxed);
      });

      while (fin_actual != 0) {
         auto ini_actual = fin_actual - std::min(fin_actual, std::uint32_t(reductores.size( )));

         tbb::parallel_for(ini_actual, fin_actual, [&, this](auto indice) {
            auto& datos = datos_[indice];
            auto& reductor = reductores[indice - ini_actual];

            if (reductor.representante.load(std::memory_order_relaxed) == &reductor.polinomio) {
               if (reductor.fin_lista.load(std::memory_order_relaxed) == nullptr) {
                  reductor.polinomio.clear(indice + 1);
               }
               else {
                  for (auto i : base::make_range(reductor.lista.get( ), reductor.fin_lista.load(std::memory_order_relaxed))) {
                     clear_segment(reductor.polinomio.vector_begin( )[i]);
                  }
               }
            }

            if (datos.reductor == std::uint32_t(-1)) {
               reductor.representante.store(nullptr, std::memory_order_relaxed);
               reductor.fin_lista.store(nullptr, std::memory_order_relaxed);
            }
            else {
               visita_producto_disperso(datos_[datos.reductor].polinomio, conv.monomial_of(indice) / conv.monomial_of(datos.reductor), sparse_assign{reductor.polinomio, indice + 1}, conv);
               reductor.representante.store(&reductor.polinomio, std::memory_order_relaxed);

               if (datos.denso) {
                  reductor.fin_lista.store(nullptr, std::memory_order_relaxed);
               }
               else {
                  reductor.fin_lista.store(this->escribe_lista_(reductor.polinomio, indice + 1, reductor.lista.get( )), std::memory_order_relaxed);
                  datos.denso = (reductor.fin_lista.load(std::memory_order_relaxed) == nullptr);
               }
            }
         });

         tbb::parallel_for_each(reducibles, [&, this](auto& reducir) {
            while (reducir.indice_lider >= ini_actual && reducir.indice_lider < fin_actual) {
               auto& datos_reducir = datos_[reducir.indice_lider];
               auto& reductor = reductores[reducir.indice_lider - ini_actual];
               const polynomial* esperado = nullptr;

               if (reductor.representante.compare_exchange_strong(esperado, &reducir.polinomio)) {
                  reductor.fin_lista.store(this->escribe_lista_(reducir.polinomio, reducir.indice_lider + 1, reductor.lista.get( )));
                  indices.push_back(reducir.indice_lider);

                  datos_reducir.clave = clave_.fetch_add(1);
                  datos_reducir.denso = (reductor.fin_lista.load(std::memory_order_relaxed) == nullptr);
                  datos_reducir.reductor = reducir.indice_lider;
                  datos_reducir.polinomio = sparse_polynomial(reducir.polinomio, reducir.indice_lider + 1);

                  return;
               }

               if (reductor.fin_lista.load( ) == nullptr) {
                  reducir.polinomio.exclusive_merge(reducir.indice_lider + 1, *reductor.representante.load(std::memory_order_relaxed));
               }
               else {
                  for (auto i : base::make_range(reductor.lista.get( ), reductor.fin_lista.load(std::memory_order_relaxed))) {
                     exclusive_merge_segment(reducir.polinomio.vector_begin( )[i], reductor.representante.load(std::memory_order_relaxed)->vector_begin( )[i]);
                  }
               }

               reducir.indice_lider = reducir.polinomio.find_leader(reducir.indice_lider);
            }
         });

         fin_actual = ini_actual;
      }

      tbb::parallel_for_each(reducibles, [&](auto& reducir) {
         reducir.polinomio.clear(reducir.indice_lider + 1);
      });
   }

   void postprocess(const tbb::concurrent_vector<std::uint32_t>& indices, tbb::concurrent_vector<std::uint32_t>& indices_eliminar, const grevlex_cache<N>& conv)
   {
      tbb::parallel_for_each(indices_eliminar, [&, this](auto indice) {
         datos_[indice].polinomio.reset( );
      });

      indices_eliminar.clear( );

      tbb::parallel_for_each(indices, [&, this](auto indice) {
         datos_[indice].reductor = std::min(indice, this->busca_reductor_(indice, conv));
      });
   }

   void diagonalize(const tbb::concurrent_vector<std::uint32_t>& indices, const grevlex_cache<N>& conv)
   {
      tbb::parallel_for(std::uint32_t(0), std::uint32_t(datos_.size( )), [&, this](auto indice) {
         auto& datos = datos_[indice];
         if (datos.polinomio.empty( )) {
            datos.reductor = this->busca_reductor_(indice, conv);
         }
      });

      std::vector<std::shared_timed_mutex> sincronizacion(datos_.size( ));
      tbb::parallel_for_each(indices, [&, this](auto indice) {
         auto& datos = datos_[indice];
         if (datos.polinomio.empty( )) {
            return;
         }

         polynomial buffer(indice + 1);
         visita_disperso(datos.polinomio, sparse_assign{buffer, indice + 1});
         auto cambios = false;

         for (std::uint32_t j = indice; j > 0; --j) {
            auto indice_interno = j - 1;
            auto& datos_interno = datos_[indice_interno];
            if (buffer.find(indice_interno) && datos_interno.reductor != std::uint32_t(-1)) {
               std::shared_lock<std::shared_timed_mutex> sinc(sincronizacion[datos_interno.reductor]);
               visita_producto_disperso(datos_[datos_interno.reductor].polinomio, conv.monomial_of(indice_interno) / conv.monomial_of(datos_interno.reductor), sparse_exclusive_merge{buffer, indice_interno + 1}, conv);

               cambios = true;
            }
         }

         if (cambios) {
            std::unique_lock<std::shared_timed_mutex> sinc(sincronizacion[indice]);
            datos.polinomio = sparse_polynomial(buffer, indice + 1);
         }
      });
   }

   tbb::concurrent_vector<std::pair<sparse_polynomial, std::uint32_t>> extract_up_to(std::uint32_t grado, const grevlex_cache<N>& conv)
   {
      tbb::concurrent_vector<std::pair<sparse_polynomial, std::uint32_t>> res;
      tbb::parallel_for(std::uint32_t(0), conv.population_up_to(grado), [&, this](auto indice) {
         auto& datos = datos_[indice];
         if (!datos.polinomio.empty( )) {
            res.emplace_back(std::move(datos.polinomio), indice);
         }
      });

      datos_.clear( );
      grado_ = std::uint32_t(-1);

      return res;
   }

private:
   std::uint32_t busca_reductor_(std::uint32_t indice, const grevlex_cache<N>& conv)
   {
      std::uint32_t res = -1;
      subset_visit(conv.monomial_of(indice), [&, this](const auto& subconjunto) {
         auto indice_subconjunto = conv.index_of(subconjunto);
         if (!datos_[indice_subconjunto].polinomio.empty( )) {
            res = std::min(res, indice_subconjunto);
         }

         return true;
      });

      return res;
   }

   std::uint32_t* escribe_lista_(const polynomial& polinomio, std::uint32_t capacidad, std::uint32_t* lista)
   {
      auto limite = lista + reductor::list_capacity(capacidad);
      auto vectores = polinomio.vectors(capacidad);

      for (std::uint32_t i = 0; i < vectores; ++i) {
         if (!empty_segment(polinomio.vector_begin( )[i])) {
            if (lista == limite) {
               return nullptr;
            }

            *lista++ = i;
         }
      }

      return lista;
   }

   std::uint32_t grado_;
   std::atomic<std::uint32_t> clave_;
   std::vector<registro> datos_;
};

namespace impl {
   template<std::uint32_t N>
   inline void reporta_estado(const piramide_polinomial<N>& piramide, const grevlex_cache<N>& conv)
   {
      std::array<std::atomic<std::uint32_t>, N + 1> stats = { };
      tbb::parallel_for(std::uint32_t(0), conv.population_up_to(piramide.degree( )), [&](auto i) {
         stats[conv.degree_of(i)] += !piramide[i].empty( );
      });

      for (std::uint32_t i = 0; i <= piramide.degree( ); ++i) {
         std::cerr << i << ": " << stats[i] << '\n';
      }

      std::cerr << '\n';
   }

   template<std::uint32_t N>
   inline void descomprime(std::atomic<typename piramide_polinomial<N>::reducible*>* escribir, const std::pair<sparse_polynomial, std::uint32_t>& comprimido, const piramide_polinomial<N>& piramide, const grevlex_cache<N>& conv)
   {
      auto actual = escribir->fetch_add(1);
      visita_disperso(comprimido.first, sparse_assign{actual->polinomio, conv.population_up_to(piramide.degree( ))});
      actual->indice_lider = comprimido.second;
   }

   template<std::uint32_t N>
   inline void multiplica(std::atomic<typename piramide_polinomial<N>::reducible*>* escribir, const std::pair<std::uint32_t, std::uint32_t>& producto, const piramide_polinomial<N>& piramide, const grevlex_cache<N>& conv)
   {
      auto actual = escribir->fetch_add(1);
      visita_producto_disperso(piramide[producto.first], conv.monomial_of(producto.second), sparse_assign{actual->polinomio, conv.population_up_to(piramide.degree( ))}, conv);
      actual->indice_lider = actual->polinomio.find_leader(conv.population_up_to(piramide.degree( )));
   }

   template<typename T>
   inline void elimina_atras(std::vector<T>& v, std::uint32_t n)
   {
      v.erase(v.end( ) - n, v.end( ));
   }

   template<typename T>
   inline void elimina_atras(tbb::concurrent_vector<T>& v, std::uint32_t n)
   {
      v.resize(v.size( ) - n);
   }
}

template<std::uint32_t N>
inline void actualiza_base(const tbb::concurrent_vector<std::pair<sparse_polynomial, std::uint32_t>>& polinomios_pendientes, tbb::concurrent_vector<std::pair<std::uint32_t, std::uint32_t>>& inducidos_pendientes, tbb::concurrent_vector<std::pair<std::uint32_t, std::uint32_t>>& multiplos_pendientes, tbb::concurrent_vector<std::uint32_t>& indices, tbb::concurrent_vector<std::uint32_t>& indices_eliminar, std::uint32_t& lineales, piramide_polinomial<N>& piramide, const grevlex_cache<N>& conv)
{
std::cerr << "preprocesando...\n";
   std::uint32_t anteriores = indices.size( );
   piramide.preprocess(multiplos_pendientes, conv);
   std::vector<std::reference_wrapper<const std::pair<sparse_polynomial, std::uint32_t>>> entrada_pendientes(polinomios_pendientes.begin( ), polinomios_pendientes.end( ));
std::cerr << "en total " << entrada_pendientes.size( ) + inducidos_pendientes.size( ) + multiplos_pendientes.size( ) << " polinomios por importar...\n";
   auto grupo_reducibles = std::min(std::size_t(2048), entrada_pendientes.size( ) + inducidos_pendientes.size( ) + multiplos_pendientes.size( ));
   auto grupo_reductores = std::min(std::size_t(48), std::size_t(piramide.size( )));

   std::vector<typename piramide_polinomial<N>::reducible> reducibles;
   for (std::uint32_t i = 0; i < grupo_reducibles; ++i) {
      reducibles.emplace_back(conv.population_up_to(piramide.degree( )));
   }

   std::vector<typename piramide_polinomial<N>::reductor> reductores;
   for (std::uint32_t i = 0; i < grupo_reductores; ++i) {
      reductores.emplace_back(conv.population_up_to(piramide.degree( )));
   }

   while (!entrada_pendientes.empty( ) || !inducidos_pendientes.empty( ) || !multiplos_pendientes.empty( )) {
      std::atomic<typename piramide_polinomial<N>::reducible*> escribir(reducibles.data( ));
      auto prepara_reducibles = [&](auto& pendientes, const auto& func, const auto&... parametros) {
         auto cuantos = std::min(grupo_reducibles - (escribir.load(std::memory_order_relaxed) - reducibles.data( )), pendientes.size( ));
         tbb::parallel_for_each(pendientes.end( ) - cuantos, pendientes.end( ), std::bind(func, &escribir, std::placeholders::_1, std::cref(parametros)...));
         impl::elimina_atras(pendientes, cuantos);
      };
impl::reporta_estado(piramide, conv);
      prepara_reducibles(entrada_pendientes, FUNCTOR(impl::descomprime<N>), piramide, conv);
      prepara_reducibles(inducidos_pendientes, FUNCTOR(impl::multiplica<N>), piramide, conv);
      prepara_reducibles(multiplos_pendientes, FUNCTOR(impl::multiplica<N>), piramide, conv);
std::cerr << "importando " << escribir.load(std::memory_order_relaxed) - reducibles.data( ) << ", restan " << entrada_pendientes.size( ) + inducidos_pendientes.size( ) + multiplos_pendientes.size( ) << '\n';
auto t0 = std::chrono::high_resolution_clock::now( );
      piramide.import(reducibles.data( ), escribir.load(std::memory_order_relaxed), reductores.begin( ), reductores.end( ), indices, conv);
auto t1 = std::chrono::high_resolution_clock::now( );
auto total = std::chrono::duration_cast<std::chrono::milliseconds>(t1 - t0).count( ) / 1000.0;
std::cerr << total / (escribir.load(std::memory_order_relaxed) - reducibles.data( )) << " segundos por reducible, total = " << total << '\n';
   }

   piramide.postprocess(indices, indices_eliminar, conv);
   lineales += std::count_if(indices.begin( ) + anteriores, indices.end( ), [&](auto indice) {
      return conv.degree_of(indice) <= 1;
   });
std::cerr << lineales << " lineales actualmente\n\n";
impl::reporta_estado(piramide, conv);
}

template<std::uint32_t N>
inline void simplifica_indices(tbb::concurrent_vector<std::uint32_t>& indices, tbb::concurrent_vector<std::uint32_t>& indices_eliminar, std::uint32_t& existentes, const piramide_polinomial<N>& piramide)
{
std::cerr << "a simplificar " << indices.size( ) << " indices\n";
   auto indices_simplificados = tbb::concurrent_vector<std::uint32_t>( );
   auto indices_eliminables = tbb::concurrent_vector<std::uint32_t>( );
   auto translado = [&](auto indice) {
      if (piramide.get_status(indice).second) {
         indices_simplificados.push_back(indice);
      }
      else {
         indices_eliminables.push_back(indice);
      }
   };

   tbb::parallel_for_each(indices.begin( ), indices.begin( ) + existentes, translado);
   auto existentes_simplificados = indices_simplificados.size( );
   tbb::parallel_for_each(indices.begin( ) + existentes, indices.end( ), translado);
std::cerr << "quedaron " << indices_simplificados.size( ) << " indices\n";
   existentes = existentes_simplificados;
   indices = std::move(indices_simplificados);
   indices_eliminar = std::move(indices_eliminables);
}

template<std::uint32_t N>
inline bool criterio_buchberger(const monomial<N>& mcm, const monomial<N>& a, std::uint32_t clave_a, const monomial<N>& b, std::uint32_t clave_b, const piramide_polinomial<N>& piramide, const grevlex_cache<N>& conv)
{
   auto datos_ab = std::make_tuple(true, mcm.degree( ), true, std::max(clave_a, clave_b), std::min(clave_a, clave_b));

   return subset_find(mcm, [&](const auto& c) {
      auto indice_c = conv.index_of(c);

      if (indice_c < piramide.size( )) {
         auto estado_c = piramide.get_status(indice_c);
         auto clave_c = estado_c.first;

         if (estado_c.second && clave_a != clave_c && clave_b != clave_c) {
            auto datos_ac = std::make_tuple(!is_coprime(a, c), least_common_multiple(a, c).degree( ), !is_divisible(b, c), std::max(clave_a, clave_c), std::min(clave_a, clave_c));
            auto datos_bc = std::make_tuple(!is_coprime(b, c), least_common_multiple(b, c).degree( ), !is_divisible(a, c), std::max(clave_b, clave_c), std::min(clave_b, clave_c));

            return std::make_pair(datos_ab > datos_ac && datos_ab > datos_bc, true);
         }
      }

      return std::make_pair(false, true);
   });
}

template<std::uint32_t N>
inline bool es_grobner(const std::array<tbb::concurrent_vector<std::pair<sparse_polynomial, std::uint32_t>>, N + 1>& polinomios_entrada, const tbb::concurrent_vector<std::uint32_t>& indices, const piramide_polinomial<N>& piramide, const grevlex_cache<N>& conv)
{
   if (piramide.get_status(0).second) {
      return true;
   }
std::cerr << "revisando si es grobner en " << piramide.degree( ) << " con " << indices.size( ) << " irreducibles\n";
   if (std::find_if(polinomios_entrada.begin( ) + piramide.degree( ) + 1, polinomios_entrada.end( ), base::not_fn(FUNCTOR(base::empty))) != polinomios_entrada.end( )) {
      return false;
   }

   tbb::task_group_context contexto;
   tbb::parallel_for(std::uint32_t(0), std::uint32_t(indices.size( )), [&](auto i) {
      auto indice1 = indices[i];
      auto lider1 = conv.monomial_of(indice1);
      auto clave1 = piramide.get_status(indice1).first;

      if (lider1.degree( ) == piramide.degree( )) {
         return (void)contexto.cancel_group_execution( );
      }

      tbb::parallel_for(std::uint32_t(0), i, [&](auto j) {
         auto indice2 = indices[j];
         auto lider2 = conv.monomial_of(indice2);
         auto clave2 = piramide.get_status(indice2).first;

         auto mcm = least_common_multiple(lider1, lider2);
         auto grado_mcm = mcm.degree( );

         if (!is_coprime(lider1, lider2) && grado_mcm > piramide.degree( ) && !criterio_buchberger(mcm, lider1, clave1, lider2, clave2, piramide, conv)) {
            return (void)contexto.cancel_group_execution( );
         }
      }, tbb::auto_partitioner( ), contexto);
   }, tbb::auto_partitioner( ), contexto);

   return !contexto.is_group_execution_cancelled( );
}

template<std::uint32_t N>
inline std::pair<tbb::concurrent_vector<std::pair<sparse_polynomial, std::uint32_t>>, bool> grobner_interrumpible(const std::array<tbb::concurrent_vector<std::pair<sparse_polynomial, std::uint32_t>>, N + 1>& polinomios_entrada, std::uint32_t lineales_esperados, grevlex_cache<N>& conv)
{
   std::uint32_t lineales = 0;
   piramide_polinomial<N> piramide;
   tbb::concurrent_vector<std::uint32_t> indices;
   tbb::concurrent_vector<std::uint32_t> indices_eliminar;

   for (std::uint32_t d = 0; d <= N; ++d) {
      conv.emplace_degree(d);
      piramide.increase_degree(conv);
std::cerr << "*** piso " << d << "...\n";
      std::uint32_t cola = 0;
      std::uint32_t existentes = indices.size( );
      tbb::concurrent_vector<std::pair<std::uint32_t, std::uint32_t>> inducidos_pendientes;
      tbb::concurrent_vector<std::pair<std::uint32_t, std::uint32_t>> multiplos_pendientes;
std::cerr << "importando entrada de " << polinomios_entrada[d].size( ) << " polinomios\n";
      actualiza_base(polinomios_entrada[d], inducidos_pendientes, multiplos_pendientes, indices, indices_eliminar, lineales, piramide, conv);

      while (cola < indices.size( ) && lineales < lineales_esperados) {
         tbb::parallel_for(cola, std::uint32_t(indices.size( )), [&](auto i) {
            auto indice1 = indices[i];
            auto lider1 = conv.monomial_of(indice1);
            auto clave1 = piramide.get_status(indice1).first;

            auto grado = lider1.degree( );
            auto grado_inducido = std::min(grado + 1, N);

            if (grado_inducido <= d && (grado_inducido == d || i >= existentes)) {
               tbb::parallel_for(std::uint32_t(0), grado, [&](auto i) {
                  inducidos_pendientes.emplace_back(indice1, conv.index_of(genera_factor<N>(lider1[i])));
               });
            }

            tbb::parallel_for(std::uint32_t(0), i, [&](auto j) {
               auto indice2 = indices[j];
               auto lider2 = conv.monomial_of(indice2);
               auto clave2 = piramide.get_status(indice2).first;

               auto mcm = least_common_multiple(lider1, lider2);
               auto grado_mcm = mcm.degree( );

               if (!is_coprime(lider1, lider2) && grado_mcm <= d && (grado_mcm == d || i >= existentes || j >= existentes) && !criterio_buchberger(mcm, lider1, clave1, lider2, clave2, piramide, conv)) {
                  multiplos_pendientes.emplace_back(indice1, conv.index_of(mcm / lider1));
                  multiplos_pendientes.emplace_back(indice2, conv.index_of(mcm / lider2));
               }
            });
         });

         simplifica_indices(indices, indices_eliminar, existentes, piramide);
         cola = indices.size( );
         actualiza_base(tbb::concurrent_vector<std::pair<sparse_polynomial, std::uint32_t>>( ), inducidos_pendientes, multiplos_pendientes, indices, indices_eliminar, lineales, piramide, conv);
      }

      if (es_grobner(polinomios_entrada, indices, piramide, conv)) {
         break;
      }

      if (lineales >= lineales_esperados) {
std::cerr << "interrumpiendo grobner por lineales\n";
         return { piramide.extract_up_to(std::min(piramide.degree( ), std::uint32_t(1)), conv), false };
      }
   }

   return { (piramide.diagonalize(indices, conv), piramide.extract_up_to(piramide.degree( ), conv)), true };
}

int main( )
{
   constexpr std::uint32_t N = 20;
   constexpr std::uint32_t D = 2;
//tbb::task_scheduler_init mt(1);
   grevlex_cache<N> conv(D);
   std::array<tbb::concurrent_vector<std::pair<sparse_polynomial, std::uint32_t>>, N + 1> polinomios_entrada;

   lectura: {
      std::uint32_t capacidad = conv.population_up_to(D);
      std::vector<monomial<N>> trabajo;
      polynomial buffer(capacidad);

      while (get_polynomial(std::cin, trabajo)) {
         lee_polinomio(buffer, trabajo.begin( ), trabajo.end( ), conv);
         auto indice_lider = buffer.find_leader(capacidad);

         if (indice_lider != std::uint32_t(-1)) {
            polinomios_entrada[conv.degree_of(indice_lider)].emplace_back(sparse_polynomial(buffer, indice_lider + 1), indice_lider);
            buffer.clear(indice_lider + 1);
         }
      }
   }
auto t0 = std::chrono::high_resolution_clock::now( );
   auto res = grobner_interrumpible(polinomios_entrada, std::ceil(std::sqrt(N)), conv);
   if (!res.second) {
      tbb::parallel_for_each(res.first, [&](auto& comprimido) {
         polinomios_entrada[conv.degree_of(comprimido.second)].push_back(std::move(comprimido));
      });

      res = grobner_interrumpible(polinomios_entrada, N + 2, conv);
   }
auto t1 = std::chrono::high_resolution_clock::now( );
std::cerr << std::chrono::duration_cast<std::chrono::milliseconds>(t1 - t0).count( ) / 1000.0 << '\n';
   escritura: {
      tbb::parallel_sort(res.first.begin( ), res.first.end( ), [&](const auto& comprimido1, const auto& comprimido2) {
         return comprimido1.second < comprimido2.second;
      });

      std::vector<monomial<N>> trabajo;
      std::for_each(res.first.rbegin( ), res.first.rend( ), [&](auto& comprimido) {
         escribe_polinomio(comprimido.first, std::back_inserter(trabajo), conv);
         std::reverse(trabajo.begin( ), trabajo.end( ));
         write_polynomial(std::cout, trabajo) << '\n';

         trabajo.clear( );
      });
   }
}
