#ifndef POLYNOMIAL_IO_H
#define POLYNOMIAL_IO_H

#include "base/bits.h"
#include <cstddef>
#include <cctype>
#include <iostream>

template<typename M>
inline std::istream& get_monomial(std::istream& is, M& m)
{
   m.clear( );

   if ((is >> std::ws).peek( ) == '1') {
      return is.ignore( );
   }

   for (;;) {
      std::size_t i;

      if (std::tolower((is >> std::ws).peek( )) == 'x' && is.ignore( ) >> i && i < m.capacity( )) {
         m.insert(i);

         if (!is.eof( ) && is >> std::ws && !is.eof( ) && is.peek( ) == '^') {
            std::size_t e;

            if (is.ignore( ) >> e && e == 0) {
               m.erase(i);
            }
         }

         if (!is.eof( ) && is >> std::ws && !is.eof( ) && is.peek( ) == '*') {
            is.ignore( );
            continue;
         }
      }
      else {
         is.setstate(std::ios_base::failbit);
      }

      break;
   }

   return is;
}

template<typename P>
inline std::istream& get_polynomial(std::istream& is, P& p)
{
   p.clear( );

   if (is >> std::ws && is.peek( ) == '0') {
      return is.ignore( );
   }

   for (;;) {
      typename P::value_type m;

      if (get_monomial(is, m)) {
         p.push_back(std::move(m));

         if (!is.eof( ) && is >> std::ws && !is.eof( ) && is.peek( ) == '+') {
            is.ignore( );
            continue;
         }
      }
      else {
         is.setstate(std::ios_base::failbit);
      }

      break;
   }

   return is;
}

template<typename M>
inline std::ostream& write_monomial(std::ostream& os, const M& m)
{
   if (m.none( )) {
      return os << '1';
   }

   auto tope = m.degree( ) - 1;

   for (std::size_t i = 0; i < tope; ++i) {
      os << 'x' << m[i] << '*';
   }

   return os << 'x' << m[tope];
}

template<typename P>
inline std::ostream& write_polynomial(std::ostream& os, const P& p)
{
   if (p.empty( )) {
      return os << '0';
   }

   auto tope = p.size( ) - 1;

   for (std::size_t i = 0; i < tope; ++i) {
      write_monomial(os, p[i]) << " + ";
   }

   return write_monomial(os, p[tope]);
}

#endif
