#ifndef SIMD_H
#define SIMD_H

#include "base/functional.h"
#include <cstddef>
#include <type_traits>
#include <x86intrin.h>

#define CACHE_LINE_SIZE 64

using block_type = unsigned long long;
using vector_type =
   #if defined(__AVX2__)
       __m256i;
   #else
      __m128i;
   #endif

static_assert(sizeof(vector_type) <= CACHE_LINE_SIZE && std::is_same<long long __attribute__((vector_size(sizeof(vector_type)))), vector_type>::value, "simd");

template<typename T>
inline bool is_zero(const T& v)
{
   return v == 0;
}

inline bool is_zero(const __m128i& v)
{
   return _mm_movemask_epi8(_mm_cmpeq_epi32(v, _mm_setzero_si128( ))) == 0xFFFF;
}

inline bool is_zero(const __m256i& v)
{
   return _mm256_testz_si256(v, v);
}

#endif
